﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using X3g.Plugin;

namespace SamplePlugin.ArticleHelper
{
    /// <summary>
    /// Interaction logic for FetchArticles.xaml
    /// </summary>
    public partial class FetchArticles : System.Windows.Window
    {
        /// For convenience use the same names as the Plugin class does
        private DocumentManager Document;
        private IArticleManager X3gArticles;

        /// We create a basket for the lifetime of this Dialog
        private IBasket basket;

        /// <summary>
        /// Initialize relevant variables in the constructor.
        /// </summary>
        public FetchArticles(DocumentManager pDocumentManager, IArticleManager pArticleManager)
        {
            Document = pDocumentManager;
            X3gArticles = pArticleManager;

            InitializeComponent();
            LoadArticlesIntoUI();
        }

        #region Methods to get Articles into the UI
        /// <summary>
        /// This function collects all Articles with the basket and assigns them to the UI#
        /// </summary>
        private void LoadArticlesIntoUI()
        {
            basket = MakeBasket();
            listBox.ItemsSource = RecursiveGetArticles(basket.RootFolder, false);
        }

        /// <summary>
        /// Simply create a Basket.
        /// Modify the Parameter to fit your needs.
        /// </summary>
        private IBasket MakeBasket()
        {
            BasketGenerationFlags flags =
                // Required for Images from Basket
                BasketGenerationFlags.PrepareCatalogImages | BasketGenerationFlags.PrepareGeneratedImages
                // Use Data from current planning
                | BasketGenerationFlags.DirectMode;
            var emptyList = new List<string>();
            return X3gArticles.GenerateBasket(flags, emptyList.AsReadOnly());
        }
        
        /// Decends all folder to get all the articles.
        /// <param name="showNonArticle">
        /// This will also attempt to show positions for "Folder" and "UnknownItem".
        /// </param>
        private List<DisplayArticle> RecursiveGetArticles(IBasketItem item, bool showNonArticle)
        {
            DisplayArticle display = null;
            var articleItem = item as IBasketArticleItem;

            if (articleItem != null)
                display = CreateDisplayArticle(articleItem);
            else if (showNonArticle)
                display = CreateDisplayArticle(item);

            var ret = new List<DisplayArticle>();
            if (display != null)
                ret.Add(display);

            foreach (var c in item.Children)
            {
                ret.AddRange(RecursiveGetArticles(c, showNonArticle));
            }

            return ret;
        }
        #endregion

        #region Methods to Select or Delete Entities
        /// <summary>
        /// Select by BasketID.
        /// Hint: You can only select Elements in model space or children of opened Blocks.
        /// </summary>
        private void SelectEntity(string basketID)
        {
            if (string.IsNullOrWhiteSpace(basketID))
                return;
            
            var entity = GetEntityForBasketId(Document.ModelSpace, basketID, false);
            if(entity == null)
            {
                var errorStr = $"BasketID {basketID} not found in ModelSpace";
                X3g.Plugin.MessageBox.Show(
                    X3g.Plugin.MessageBox.Icon.Error,
                    errorStr);
                return;
            }

            Document.SetSelection(new List<DbEntity>() { entity });
        }

        private DbEntity GetEntityForBasketId(X3g.Plugin.Block block, string basketID, bool recursive)
        {
            if (block == null)
                return null;

            // First we check if our Entity is in the current Block
            foreach (var entity in block.GetEntities())
            {
                if (CompareEntityWithBasketID(entity, basketID))
                    return entity;
            }

            // If desired, we can now look deeper into the tree
            if (!recursive)
                return null;
            
            foreach (var blockRef in block.GetEntities<DbBlockReference>())
            {
                var childBlock = Document.Blocks.GetBlock(blockRef.Id);
                var ret = GetEntityForBasketId(childBlock, basketID, recursive);
                if (ret != null)
                    return ret;
            }

            return null;
        }

        /// <summary>
        /// Returns true if the given DbEntity is listed in the the basked with the given BasketID.
        /// </summary>
        private bool CompareEntityWithBasketID(DbEntity entity, string basketID)
        {
            var entityId = entity.Id;
            if (string.IsNullOrWhiteSpace(entityId))
                return false;

            if(X3gArticles.IsArticle(entityId))
            {
                var article = X3gArticles.GetArticleEntity(entityId);
                if (article?.GetInformation()?.BasketId == basketID)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Get informations from the IBasketArticleItem and place it int a new DisplayArticle for the UI.
        /// </summary>
        private DisplayArticle CreateDisplayArticle(IBasketArticleItem item)
        {
            var bitmap = new BitmapImage();
            if (!string.IsNullOrWhiteSpace(item?.ArticleImage))
            {
                /// We have to tell the system that out image source
                /// is not to be Cached. The default Cache option of 
                /// BitmapImage keeps a file-pointer open, which 
                /// prevents any cleanup of the temporary Basket 
                /// folder. (Thus causing Exceptions)
                bitmap.BeginInit();
                bitmap.CacheOption = BitmapCacheOption.OnLoad;
                bitmap.CreateOptions = BitmapCreateOptions.IgnoreImageCache;
                bitmap.UriSource = new Uri(item.ArticleImage);
                bitmap.EndInit();
            }

            var amount = item.Calculation?.Quantity ?? 0;
            // grossPrice might be 'null' if article contains errors
            var grossPrice = item.Calculation?.GrossPrice;
            var priceStr = (grossPrice?.Single?.Value ?? 0).ToString("N");
            var unit = grossPrice?.Single?.Unit;
            if (unit != null)
            {
                priceStr += " (";
                priceStr += unit;
                priceStr += ")";
            }

            return new DisplayArticle()
            {
                Position = item.Position,
                ImageSrc = bitmap,
                DisplayText = item.GetDescription(BasketArtDescrType.Default, basket?.AvailableLanguages?.First()),
                PricePerPiece = priceStr,
                Amount = amount.ToString("N"),
                BasketID = item.Id
            };
        }

        /// <summary>
        /// If the other CreateDisplayArticle does not work (e.g.: for Folder or Unknown Types) this
        /// overloaded function will attempt to create something for a BasketItem.
        /// </summary>
        private DisplayArticle CreateDisplayArticle(IBasketItem item)
        {
            return new DisplayArticle()
            {
                Position = item.Position,
                ImageSrc = null,
                DisplayText = item.GetLabel(basket?.AvailableLanguages?.First()),
                PricePerPiece = string.Empty,
                Amount = 1.ToString("N"),
                BasketID = item.Id
            };
        }
        #endregion

        #region Events
        private void SelectButton_Click(object sender, RoutedEventArgs e)
        {
            var selectedItem = listBox.SelectedItem as DisplayArticle;
            SelectEntity(selectedItem?.BasketID);
        }
        #endregion

        #region helper class
        /// <summary>
        /// A container to fill the ListBox.
        /// </summary>
        public class DisplayArticle
        {
            public string Position { get; set; } = "";
            public ImageSource ImageSrc { get; set; }
            public string DisplayText { get; set; } = "";
            public string PricePerPiece { get; set; } = "";
            public string Amount { get; set; } = "0";
            public string BasketID { get; set; } = null;
        }
        #endregion
    }
}
