﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using X3g.Plugin;

namespace SamplePlugin.CustomEntity
{
    enum GridStyle : int
    {
        Lines = 0,
        Points = 1
    }

    class Grid : IPropertyCallbacks
    {
        const string CustomType = "egr::SamplePlugin::Grid";

        //property keys
        const string GridPropertyGroupKey = "egr::SamplePlugin::Grid::PropGroup";
        const string ColumnsPropertyKey = "egr::SamplePlugin::Grid::Columns";
        const string RowsPropertyKey = "egr::SamplePlugin::Grid::Rows";
        const string StylePropertyKey = "egr::SamplePlugin::Grid::Style";

        DbBlockReference gridEntity;
        DocumentManager parentDocument;

        int Columns
        {
            get
            {
                DbDictionary sampleDict = gridEntity.CustomData.GetDictionary("SamplePlugin", false);
                if (sampleDict == null)
                    return (0);

                int numColumns = 0;
                sampleDict.Get(ColumnsPropertyKey, out numColumns);
                return (numColumns);
            }

            set
            {
                DbDictionary sampleDict = gridEntity.CustomData.GetDictionary("SamplePlugin", true);                
                if (value != this.Columns)
                {
                    sampleDict.Set(ColumnsPropertyKey, value);
                    //XXX update geometry
                    RedrawGrid();
                }
            }
        }

        int Rows
        {
            get
            {
                DbDictionary sampleDict = gridEntity.CustomData.GetDictionary("SamplePlugin", false);
                if (sampleDict == null)
                    return (0);

                int numRows = 0;
                sampleDict.Get(RowsPropertyKey, out numRows);
                return (numRows);
            }

            set
            {
                DbDictionary sampleDict = gridEntity.CustomData.GetDictionary("SamplePlugin", true);
                if (value != this.Rows)
                {
                    sampleDict.Set(RowsPropertyKey, value);
                    //XXX update geometry
                    RedrawGrid();
                }
            }
        }

        GridStyle Style
        {
            get
            {
                int intValue;
                DbDictionary sampleDict = gridEntity.CustomData.GetDictionary("SamplePlugin", false);
                if (sampleDict != null && sampleDict.Get(StylePropertyKey, out intValue))
                {
                    return (GridStyle)intValue;
                }
                else
                {
                    return GridStyle.Lines; //default
                }
            }
            set
            {
                if (Style == value)
                    return;
                DbDictionary sampleDict = gridEntity.CustomData.GetDictionary("SamplePlugin", true);
                if (value != this.Style)
                {
                    sampleDict.Set(StylePropertyKey, (int)value);
                    //XXX update geometry
                    RedrawGrid();
                }
            }
        }

        public Grid(DocumentManager document, GeBoundingBox3d bounds, int rows, int columns)
        {
            double gridWidth = bounds.Max.x - bounds.Min.x;
            double gridHeight = bounds.Max.y - bounds.Min.y;

            parentDocument = document;
            //create a new anonymous block
            Block gridBlock = document.Blocks.CreateBlock("*E");

            //create row lines within the block
            double rowHeight = gridHeight / rows;
            for (int r = 0; r < rows + 1; r++)
            {
                DbPolyline rowLine = new DbPolyline(gridBlock);
                rowLine.SetVertices(new[] {
                    new GeVec3d(0, r * rowHeight, 0),
                    new GeVec3d(gridWidth, r * rowHeight, 0) });
            }

            //create column lines within the block
            double columnWidth = gridWidth / columns;
            for (int c = 0; c < columns + 1; c++)
            {
                DbPolyline columnLine = new DbPolyline(gridBlock);
                columnLine.SetVertices(new[] {
                    new GeVec3d(c * columnWidth, 0, 0),
                    new GeVec3d(c * columnWidth, gridHeight, 0) });
            }

            //create grid reference in modelspace
            gridEntity = new DbBlockReference(document.ModelSpace, gridBlock);
            gridEntity.CustomType = Grid.CustomType;
            gridEntity.Position = bounds.Min;
            gridEntity.UpdateGeometry();

            //as this values are attached to the entity
            //we must not set them before the entity is created
            this.Rows = rows;
            this.Columns = columns;

            //attach properties
            gridEntity.GetPropertyProvider().Add(this);
        }

        // implementation of IPropertyCallbacks

        public IList<IPropertyDescriptor> GetProperties()
        {
            var props = new List<IPropertyDescriptor>();

            PropertyDescriptor propGroup = new PropertyDescriptor(
                GridPropertyGroupKey, PropertyType.Group, I18N._("Grid"), "");
            props.Add(propGroup);

            propGroup.AddSubProperty(new PropertyDescriptor(ColumnsPropertyKey, PropertyType.Int, I18N._("Columns"), I18N._("Number of columns")));
            propGroup.AddSubProperty(new PropertyDescriptor(RowsPropertyKey, PropertyType.Int, I18N._("Rows"), I18N._("Number of rows")));

            PropertyDescriptor stylePropDesc = new PropertyDescriptor(
                StylePropertyKey, PropertyType.Choice, I18N._("Style"), I18N._("Select grid style."));
            foreach (GridStyle style in Enum.GetValues(typeof(GridStyle)))
            {
                stylePropDesc.AddChoiceValue(new ChoicePropertyValue(style.ToString(), style.ToString()));
            }
            propGroup.AddSubProperty(stylePropDesc);

            return props;
        }

        public PropertyState GetPropertyState(string pKey)
        {
            return PropertyState.Enabled;
        }

        public PropertyValue GetPropertyValue(string pKey)
        {
            if (pKey == ColumnsPropertyKey)
            {
                return new IntPropertyValue(this.Columns);
            }
            else if (pKey == RowsPropertyKey)
            {
                return new IntPropertyValue(this.Rows);
            }
            else if (pKey == StylePropertyKey)
            {
                return new ChoicePropertyValue(Style.ToString(), Style.ToString());
            }
            return null;
        }

        public void SetPropertyValue(string pKey, PropertyValue pValue)
        {
            if (pKey == ColumnsPropertyKey)
            {
                this.Columns = (pValue as IntPropertyValue).Value;
            }
            else if (pKey == RowsPropertyKey)
            {
                this.Rows = (pValue as IntPropertyValue).Value;
            }
            else if (pKey == StylePropertyKey)
            {
                this.Style = (GridStyle)Enum.Parse(typeof(GridStyle), (pValue as ChoicePropertyValue).Key);
            }
        }

        /// <summary>
        /// called when ros or columns or style changed. Deletes the lines, points of the grid and creates new lines, points
        /// </summary>
        private void RedrawGrid()
        {
            Block block = parentDocument.Blocks.GetBlock(gridEntity.Block);
            GeBoundingBox3d bounds = gridEntity.GetBoundingBox(GeMatrix.Identity);
            foreach (var item in block.GetEntities<DbPolyline>())
            {
                item.Delete();
            }
            foreach (var item in block.GetEntities<DbPoint>())
            {
                item.Delete();
            }
            //create row lines within the block
            GeVec3d diag = bounds.Max - bounds.Min;
            double rowHeight = diag.y / this.Rows;
            double columnWidth = diag.x / this.Columns;

            if (this.Style == GridStyle.Lines)
            {
                for (int r = 0; r < this.Rows + 1; r++)
                {
                    DbPolyline rowLine = new DbPolyline(block);
                    rowLine.SetVertices(new[] {
                    new GeVec3d(0, r * rowHeight, 0),
                   new GeVec3d(diag.x, r * rowHeight, 0) });
                }
             
                for (int c = 0; c < this.Columns + 1; c++)
                {
                    DbPolyline columnLine = new DbPolyline(block);
                    columnLine.SetVertices(new[] {
                    new GeVec3d(c * columnWidth, 0, 0),
                    new GeVec3d(c * columnWidth, diag.y, 0) });
                }
            }else
            {
                for (int xx = 0; xx < this.Columns; xx++)
                {
                    for (int yy = 0; yy < this.Rows; yy++)
                    {
                        DbPoint point = new DbPoint(block);
                        point.Position = new GeVec3d((xx+0.5) * columnWidth, (yy+0.5) * rowHeight, 0);
                    }
                }
            }
            gridEntity.UpdateGeometry();
        }
    }
}
